import time
import pytest

import numpy as np
from aestream import Event, FileInput

from . import _has_cuda_torch, _has_torch


class namespace:
    pass


def test_load_aedat4():
    f = FileInput("example/sample.aedat4", shape=(600, 400))
    buf = f.load()

    assert len(buf) == 117667
    assert buf[0]["timestamp"] == 1633953690975950
    assert buf[0]["x"] == 218
    assert buf[0]["y"] == 15
    assert buf[0]["polarity"] == True


def test_stream_aedat4():
    with FileInput(
        filename="example/sample.aedat4", shape=(346, 260), ignore_time=True
    ) as stream:
        time.sleep(0.3)
        interval = 0.5
        t_0 = time.time()
        events = 0
        while True:
            if time.time() > t_0 + interval:
                break
            frame = stream.read()
            if _has_torch():
                import torch

                assert isinstance(frame, torch.Tensor)
            else:
                assert isinstance(frame, np.ndarray)
            events += frame.sum()
    assert events == 117667


def test_load_dat():
    f = FileInput("example/sample.dat", shape=(600, 400))
    buf = f.load()

    assert len(buf) == 539481
    assert buf[0]["timestamp"] == 0
    assert buf[0]["x"] == 237
    assert buf[0]["y"] == 121
    assert buf[0]["polarity"] == True


def test_stream_dat():
    with FileInput(
        filename="example/sample.dat", shape=(600, 400), ignore_time=True
    ) as stream:
        time.sleep(0.3)
        interval = 0.5
        t_0 = time.time()
        events = 0
        while True:
            if time.time() > t_0 + interval:
                break
            frame = stream.read()
            if _has_torch():
                import torch

                assert isinstance(frame, torch.Tensor)
            else:
                assert isinstance(frame, np.ndarray)
            events += frame.sum()
    assert events == 539481


@pytest.mark.skipif(not _has_cuda_torch(), reason="Torch-gpu is not installed")
def test_stream_dat_torch_cuda():
    import torch

    with FileInput(
        filename="example/sample.dat", shape=(600, 400), ignore_time=True, device="cuda"
    ) as stream:
        time.sleep(1)
        interval = 0.4
        t_0 = time.time()
        events = 0
        while True:
            if time.time() > t_0 + interval:
                break
            frame = stream.read()
            assert isinstance(frame, torch.Tensor)
            assert frame.device.type == "cuda"
            events += frame.sum()
            time.sleep(0.1)
        events += stream.read().sum()
    assert events == 539481



"""
AEStream library for streaming address-event representations.

Please refer to https://github.com/aestream/aestream for usage
"""
import logging

try:
    import torch
except ImportError:
    logging.debug("Failed to import Torch: AEStream is running in Numpy mode")

# Import AEStream modules
from aestream.aestream_ext import Event
from aestream._input import FileInput, UDPInput

modules = []
try:
    from aestream._input import USBInput

    modules.append("USBInput")
except ImportError:
    logging.debug("Failed to import AEStream USB Input")

try:
    import aestream._genn as genn

    modules.append("genn")
except ImportError as ex:
    logging.debug("Failed to import GeNN: AEStream cannot use GeNN device")

try:
    from aestream._input import SpeckInput

    modules.append("SpeckInput")
except ImportError:
    logging.debug("Failed to import ZMQ: AEStream cannot use ZMQ input")

del logging

__all__ = ["Event", "FileInput", "UDPInput"] + modules
del modules



def _has_torch():
    try:
        import torch

        return True
    except ImportError:
        pass
    return False


def _has_cuda_torch():
    try:
        import torch

        return torch.has_cuda
    except ImportError:
        pass
    return False



import numpy as np
from pygenn.genn_model import create_custom_neuron_class

genn_input_model = create_custom_neuron_class(
    "genn_input",
    extra_global_params=[("input", "uint32_t*")],
    threshold_condition_code="""
    $(input)[$(id) / 32] & (1 << ($(id) % 32))
    """,
    is_auto_refractory_required=False,
)


def add_input(model, name, resolution):
    # Calculate total number of neurons in input
    # and hence required words to encode
    num_neurons = np.prod(resolution)
    num_words = (num_neurons + 31) // 32

    # Add input and initialise extra global parameter
    pop = model.add_neuron_population("input", num_neurons, genn_input_model, {}, {})
    pop.set_extra_global_param("input", np.empty(num_words, dtype=np.uint32))

    return pop



from aestream import aestream_ext as ext

# Set Numpy Event dtype
try:
    import numpy as np

    event_dtype = np.dtype(("=V", 13))
    NUMPY_EVENT_DTYPE = np.dtype(
        (
            event_dtype,
            {
                "timestamp": (np.int64, 0),
                "x": (np.int16, 8),
                "y": (np.int16, 10),
                "polarity": (np.byte, 12),
            },
        )
    )

except ImportError as e:
    raise ImportError("Numpy is required but could not be imported", e)

try:
    import torch

    USE_TORCH = True
except ImportError:
    USE_TORCH = False


class FileInput(ext.FileInput):
    def load(self):
        buffer = self.load_all()
        return np.frombuffer(buffer.data, NUMPY_EVENT_DTYPE)

    def read(self):
        t = self.read_buffer()
        if USE_TORCH:
            return t.to_torch()
        else:
            return t.to_numpy()

    def read_genn(self, population):
        # **YUCK** I would like to implement this with a mixin
        # to reduce copy-paste but seemingly nanobind doesn't like this
        # Read from stream into GeNN-owned memory
        super().read_genn(population.extra_global_params["input"].view)

        # Copy data to device
        # **NOTE** this may be a NOP if CPU backend is used
        population.push_extra_global_param_to_device("input")

        return population.extra_global_params["input"].view


class UDPInput(ext.UDPInput):
    def read(self):
        t = self.read_buffer()
        if USE_TORCH:
            return t.to_torch()
        else:
            return t.to_numpy()

    def read_genn(self, population):
        # **YUCK** I would like to implement this with a mixin
        # to reduce copy-paste but seemingly nanobind doesn't like this
        # Read from stream into GeNN-owned memory
        super().read_genn(population.extra_global_params["input"].view)

        # Copy data to device
        # **NOTE** this may be a NOP if CPU backend is used
        population.push_extra_global_param_to_device("input")

        return population.extra_global_params["input"].view


try:

    class USBInput(ext.USBInput):
        def read(self):
            t = self.read_buffer()
            if USE_TORCH:
                return t.to_torch()
            else:
                return t.to_numpy()

        def read_genn(self, population):
            # **YUCK** I would like to implement this with a mixin
            # to reduce copy-paste but seemingly nanobind doesn't like this
            # Read from stream into GeNN-owned memory
            super().read_genn(population.extra_global_params["input"].view)

            # Copy data to device
            # **NOTE** this may be a NOP if CPU backend is used
            population.push_extra_global_param_to_device("input")

            return population.extra_global_params["input"].view

except:
    pass  # Ignore if drivers are not installed

try:

    class SpeckInput(ext.SpeckInput):
        def read(self):
            t = self.read_buffer()
            if USE_TORCH:
                return t.to_torch()
            else:
                return t.to_numpy()

        def read_genn(self, population):
            # **YUCK** I would like to implement this with a mixin
            # to reduce copy-paste but seemingly nanobind doesn't like this
            # Read from stream into GeNN-owned memory
            super().read_genn(population.extra_global_params["input"].view)

            # Copy data to device
            # **NOTE** this may be a NOP if CPU backend is used
            population.push_extra_global_param_to_device("input")

            return population.extra_global_params["input"].view

except Exception as e:
    print("NO SPECK", e)
    pass  # Ignore if Speck/ZMQ isn't installed



# AEStream API Prototyping

from typing import Union, Tuple
import numpy as np

try:
    import torch
except:
    "torch not found, proceeding with numpy only"

class FileInput:
    """
    If you see this, its good :)
    """

    def init(): ...
    def read():
        """
        Read :P
        """
        ...

class FileInputReader:
    """
    Load or Iterate Events or Frames from a file

    Supported file formats: aedat, aedat4, dat

    Args:
        file_path (str): Path of file to be used
        resolution (int, int): X,Y resolution of file

    Attributes:
        file_path (str): Path of file to be used
        resolution (int, int): X,Y resolution of file
    """

    def __init__(self, file_path: str, resolution: Tuple[int, int]) -> None: ...
    def load(
        self, frame_ms: int = 0, device: str = "cpu"
    ) -> Union[np.ndarray, torch.tensor]:
        """
        Loads the whole file onto device memory

        Args:
            frame_ms (int): Selects data format. When `0`, events are loaded as is.
                Otherwise, events are accumulated into frames with synchronous period of `frame_ms`
            device (str: `cpu`, `cuda`): Selects the device to which the file is loaded

        Returns:
            Array/Tensor containing the contents of the file
        """
        ...
    def get_iterator(
        self,
        frame_ms: int = 0,
        device: str = "cpu",
        mode: str = "safe",
        memory_limit: int = 1024,
    ) -> FileInputIterator:
        """
        Provides functionality that allows files larger than device memory to be iteratively loaded

        Args:
            frame_ms (int): Selects data format. When `0`, events are loaded as is. Otherwise, events are accumulated into frames with synchronous period of `frame_ms`
            device (str: `cpu`, `cuda`): Selects the device to which the file is loaded.
            mode (str: `safe`, `unsafe`): Selects if slower deterministic (`safe`) or faster nondeterministic (`unsafe`) method is used to create frames. When using `unsafe` the exact frame a event is part of is subject to jitter at runtime.

        Returns:
            A generator that loads events or frames to `device` as specified.
        """
        ...

class FileInputIterator:
    """
    Load or Iterate Events or Frames from a file

    Supported file formats: aedat, aedat4, dat

    Args:
        file_path (str): Path of file to be used
        resolution (int, int): X,Y resolution of file

    Attributes:
        file_path (str): Path of file to be used
        resolution (int, int): X,Y resolution of file
    """

    def __init__(self, file_path: str, resolution: Tuple[int, int]) -> None: ...



import os
from skbuild import setup

pwd = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(pwd, "README.md"), encoding="utf-8") as fp:
    readme_text = fp.read()

# C++ config
cmake_args = ["-DUSE_PYTHON=1"]

# Activate CUDA, if found
try:
    import torch
    from torch.utils import cpp_extension

    cuda_home = cpp_extension._find_cuda_home()
    if cuda_home is not None:
        flags = " ".join(cpp_extension._get_cuda_arch_flags())
        cmake_args += [
            f"-DUSE_CUDA=1",
            f"-DCMAKE_CUDA_FLAGS={flags}",
            f"-DCMAKE_CUDA_COMPILER={cuda_home}/bin/nvcc",
            f"-DCUDA_INCLUDE_DIRS={cuda_home}/include",
        ]
except:
    pass

# Setuptools entrypoint
setup(
    name="aestream",
    version="0.6.0",
    author="Jens E. Pedersen",
    author_email="jens@jepedersen.dk",
    url="https://github.com/aestream/aestream",
    description="Streaming library for Address-Event Representation (AER) data",
    license="MIT",
    long_description=readme_text,
    long_description_content_type="text/markdown",
    python_requires=">=3.7",
    packages=["aestream"],
    install_requires=["numpy", "pysdl2-dll"],
    extras_require={"torch": ["torch"]},
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: C++",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Hardware :: Universal Serial Bus (USB)",
    ],
    cmake_args=cmake_args,
    package_data={"aestream": ["*.pyi", "*.typed"]},
)



import multiprocessing
import socket
import time
import pytest

import numpy
from aestream import UDPInput

from . import _has_cuda_torch


def stream_fake_data(port):
    data = b"\x0F\x00\xDA\x00"
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.sendto(data, ("127.0.0.1", port))
    sock.close()


def start_stream(port):
    p = multiprocessing.Process(target=stream_fake_data, args=(port,))
    p.start()
    return p


def test_udp():
    with UDPInput((640, 480), device="cpu") as stream:
        start_stream(3333)  # Start streaming from file

        interval = 0.5
        t_0 = time.time()
        while True:
            if t_0 + interval <= time.time():
                frame = stream.read()
                break
    assert numpy.equal(frame[218, 15], 1)


@pytest.mark.skipif(not _has_cuda_torch(), reason="Torch-gpu is not installed")
def test_udp_gpu():
    import torch

    if torch.has_cuda:
        with UDPInput((640, 480), device="cuda", port=3334) as stream:
            start_stream(3334)  # Start streaming from file

            interval = 0.5
            time.sleep(0.5)
            t_0 = time.time()
            while True:
                if t_0 + interval <= time.time():
                    frame = stream.read()
                    print(frame.argmax(0).argmax(), frame.argmax(1).argmax())
                    break

        assert torch.eq(frame[218, 15], 1)


if __name__ == "__main__":
    test_udp()
    test_udp_gpu()
